%% [Tutorial T11.2b] QSpace: NRG: Impurity spectral function
% <https://us02web.zoom.us/j/4363850698 Zoom session>, 30.06.2020
% 
% Author: <https://www.theorie.physik.uni-muenchen.de/lsvondelft/members/sci_mem/seung_sup-lee/index.html 
% Seung-Sup Lee>
% 
% Date: 20.06.2020  (Minor update: 30.06.2020)
%% 
% Here we will compute the *thermal expectation value of impurity operators* 
% and the *impurity spectral function,* by using the *full-density-matrix NRG* 
% (fdmNRG).
% 
% Consider an operator, say $\hat{O}_d$, that acts on the impurity. Then the 
% spectral function of the retarded-time correlation function of $\hat{O}_d$ is 
% given by:
% 
% $$A(\omega) = \frac{-1}{\pi} \mathrm{Im} \int \mathrm{d} t \,\, e^{i\omega 
% t} \left( -i \Theta (t) \langle [ \hat{O}_d (t), \hat{O}_d^\dagger ]_\pm \rangle 
% \right) , \qquad \qquad (1)$$
% 
% where the superscript $\pm$ of the commutator means the anti-commutator $[ 
% \,\, , \,\, ]_+$ for fermionic $\hat{O}_d$ and the commutator $[ \,\, , \,\, 
% ]_-$ for bosonic $\hat{O}_d$. With the energy eigenstates $\{ | \alpha \rangle 
% \}$ and their corresponding eigenvalues $\{ E_\alpha \}$, we write the *Lehmann 
% representation* of the spectral function as
% 
% $$A(\omega) = \sum_{\alpha \beta} \langle \alpha | \hat{O}_d | \beta \rangle 
% \left( \langle \beta | \hat{O}_d^\dagger \, \hat{\rho} | \alpha \rangle \pm 
% \langle \beta | \hat{\rho} \, \hat{O}_d^\dagger | \alpha \rangle \right) \,\, 
% \delta \left(\omega - (E_\beta - E_\alpha) \right) ,$$
% 
% where $\hat{\rho}$ is the density matrix of the system at thermal equilbrium,
% 
% $$\langle \alpha | \hat{\rho} | \beta \rangle = \delta^{\alpha}_{\,\,\, \beta} 
% \frac{\exp (-E_\beta /T)}{Z} ,$$
% 
% and $Z = \sum_\alpha \exp (- E_\alpha / T)$ is the partition function. In 
% this tutorial, we will use $\beta$ as an index, not as the inverse temperature.
% 
% As explained in the lecture, the discarded eigenstates in the iterative diagonalization 
% consitute the *complete basis of (approximate) energy eigenstates* $\{ | \alpha, 
% e \rangle_\ell^D \}$. However, it is hard to compute the spectral weight associated 
% with $| \alpha, e \rangle_\ell^D$ and $| \beta, e' \rangle_{\ell'}^D$ that are 
% at different iterations. Thus we employ the *NRG approximation* to compute spectral 
% functions:
% 
% $$A(\omega) = \sum_\ell \sum_{X, Y \neq K K} \sum_{\alpha \beta} A_{\alpha 
% \beta}^{\ell X Y}\,\, \delta \left(\omega - (E_{\beta}^{\ell Y}  - E_{\alpha}^{\ell 
% X}) \right) \, ,$$
% 
% $$A_{\alpha \beta}^{\ell X Y} = \left[ B^{\quad \, X}_{[\ell] \,\,\, Y} \right]^{\alpha}_{\,\,\, 
% \beta} \left( \left[ C^{\quad \,\, Y}_{[\ell]\,\,\, X} \rho_{[\ell] \,\,\, X}^{\quad 
% X} \right]^\beta_{\,\,\, \alpha} \pm \left[ \rho_{[\ell] \,\,\, Y}^{\quad Y} 
% C^{\quad \,\, Y}_{[\ell]\,\,\, X}  \right]^\beta_{\,\,\, \alpha} \Big) \, ,$$
% 
% $$\left[ B^{\quad \, X}_{[\ell] \,\,\, Y} \right]^{\alpha}_{\,\,\, \beta} 
% = {}_\ell^X \langle \alpha | \hat{O}_d | \beta \rangle_\ell^Y \, ,$$
% 
% $$\left[ C^{\quad \, Y}_{[\ell] \,\,\, X} \right]^{\beta}_{\,\,\, \alpha} 
% = {}_\ell^Y \langle \beta | \hat{O}_d^\dagger | \alpha \rangle_\ell^X = \left( 
% \left[ B^{\quad \, X}_{[\ell] \,\,\, Y} \right]^{\alpha}_{\,\,\, \beta} \right)^* 
% \, .$$
% 
% Here $\ell$ is the iteration index, and $X$ and $Y$ are either $K$ (kept) 
% or $D$ (discarded). The case $X = Y = K$ is not considered to avoid double-counting.
% 
% The density matrix in the discarded space is given by the diagonal matrix,
% 
% $$[\rho_{[\ell] \,\, D}^{\quad \, D}]^{\alpha}_{\,\,\, \beta}= \sum_e {}_\ell^D 
% \langle \alpha, e | \hat{\rho} | \beta, e \rangle_\ell^D= \delta^{\alpha}_{\,\,\, 
% \beta} \frac{d^{N - \ell} \exp (-E_{\alpha}^{\ell D} /T)}{Z} ,$$
% 
% where $d$ is the dimension of the local Hilbert space of each chain site, 
% and $N$ is the index of the last chain site. In the construction of the complete 
% basis $\{ | \alpha, e \rangle_{\ell}^D \} = \{ | \alpha \rangle_{\ell}^D \otimes 
% | \sigma_{\ell + 1} \rangle \otimes | \sigma_{\ell + 2} \rangle \otimes \cdots 
% \otimes | \sigma_{N} \rangle \}$, the "environmental states" $| e \rangle = 
% | \sigma_{\ell + 1} \rangle \otimes | \sigma_{\ell + 2} \rangle \otimes \cdots 
% \otimes | \sigma_{N} \rangle$ does not contribute to the energy; they only introduce 
% the degeneracy. So $d^{N-\ell}$ counts the degeneracy introduced by the environmental 
% states.
% 
% On the other hand, the density matrix in the kept space is given by tracing 
% out the Hilbert spaces of later chain sites,
% 
% $$\rho_{[\ell] \,\, K}^{\quad \, K} = \sum_e {}_\ell^K \langle \alpha, e | 
% \hat{\rho} | \beta, e \rangle_\ell^K$$
% 
% $$= \sum_{m > \ell} \sum_{\vec{\sigma}} \sum_{\alpha} \langle \sigma_{\ell 
% + 1} | \cdots \langle \sigma_{m} | \alpha \rangle_m^D \,\,\, \left[ \rho_{[m] 
% \,\, D}^{\quad \, D} \right]^\alpha_{\,\,\, \alpha} \,\,\, {}_m^D \langle \alpha 
% | \sigma_{m} \rangle \cdots | \sigma_{\ell+1} \rangle .$$
% 
% And there are no off-diagonal blocks, i.e., $\rho_{[\ell] \,\, X}^{\quad \, 
% Y} = 0$ if $X \neq Y$. The name of the method, "full density matrix", is given 
% by the fact that the density matrices from the *all* iterations (i.e., all energy 
% scales) are considered.
% 
% The Lehmann representation above represents the discrete spectral data of 
% the spectral function; the result is the collection of the Dirac delta functions, 
% rather than continuous curves. This discreteness originates from the logarithimic 
% discretization of the originally continuous system. To return to the original 
% continuity, we need to broaden the discrete data to obtain the continuous curves. 
% For this, we use the broadening kernel that is the convolution of the symmetric 
% logarithmic Gaussian (for broadening for high frequencies) and the derivative 
% of the Fermi-Dirac distribution function (for broadening for low frequencies), 
% as developed in [<https://journals.aps.org/prb/abstract/10.1103/PhysRevB.94.235127 
% S.-S. B. Lee and A. Weichselbaum, Phys. Rev. B *94*, 235127 (2016)> or <https://arxiv.org/abs/1609.01248 
% its arXiv version>].
% 
% Please review the following newly added functions:
%% 
% * |getRhoFDMQS_Ex.m|: It is zipped together with this document. This function 
% is designed to construct the density matrices $\rho_{[\ell] \,\, D}^{\quad \, 
% D}$ and $\rho_{[\ell] \,\, K}^{\quad \, K}$. *This function should be completed, 
% as Exercise 1.* See below for details.
% * |getAdiscQS_Ex.m|: It is zipped together with this document. This function 
% computes the Lehmann representation of the spectral function. Its second input 
% corresponds to $\hat{O}_d$ above. The first output |odisc| is the frequency 
% grid, and the second output |Adisc| is the histogram of the spectral weights 
% $A_{\alpha \beta}^{\ell X Y}$ according to the frequency grid. That is, the 
% value of |Adisc(n)| is the sum of all the spectral weights $A_{\alpha \beta}^{\ell 
% X Y}$ whose corresponding frequency values $E_{\beta}^{\ell Y}  - E_{\alpha}^{\ell 
% X}$ are close to the frequency bin value |odisc(n)|. *This function should be 
% completed, as Exercise 2.* See below for details.
% * |NRG/getAcont.m|: Broaden the discrete data to obtain continuous curves. 
% In this tutorial, we will not discuss the detail of the function. So take this 
% function as granted. Interested readers may refer to [<https://journals.aps.org/prb/abstract/10.1103/PhysRevB.94.235127 
% S.-S. B. Lee and A. Weichselbaum, Phys. Rev. B *94*, 235127 (2016)> or <https://arxiv.org/abs/1609.01248 
% its arXiv version>] for details. _(Yes, it's a shameless self-advertisement...)_
%% Iterative diagonalization
% First, run the iterative diagonalization for the SIAM. Here we use the same 
% parameters as those used in the last tutorial.

clear

% Hamiltonian parameters
U = 4e-3; % Coulomb interaction at the impurity
epsd = -U/2; % impurity on-site energy
Gamma = 8e-5*pi; % hybridization strength

% NRG parameters
Lambda = 2.5; % discretization parameter
N = 55; % length of the Wilson chain
Nkeep = 300;

% Wilson chain
[ff,gg] = doCLD([-1 1],[1 1]*Gamma/pi,Lambda,N);
%% 
% Note that we don't need to post-process the Wilson chain parameters as we 
% did for the tutorial T11a, since the calculations in this tutorial are stable 
% against the (very) small numerical noise that mattered in T11a.

% symmetries
symstr = 'Acharge,SU2spin'; % U(1) charge and SU(2) spin
% symstr = 'Acharge,Aspin'; % U(1) charge and U(1) spin

% Construct local operators
[F,Z,S,I] = getLocalSpace('FermionS',symstr,'NC',1);
[F,Z,S,EF] = setItag('s00','op',F,Z,S,I.E);

% particle number operator
NF = QSpace;
for itF = 1:numel(F)
    NF(itF) = contract(F(itF),'!2*',F(itF));
end
       
% Impurity Hamiltonian
H0 = U/2*sum(NF)*(sum(NF)-1) + epsd*sum(NF) + 1e-33*EF;

% ket tensor for the impurity
A0 = getIdentity(setItag('L00',getvac(EF,1)),1,EF,1,'K00',[1,3,2]);

H0 = contract(A0,'!2*',{A0,H0});

% same hopping amplitude and on-site energies for all flavors
ff = repmat(ff,[1,numel(F)]);
gg = repmat(gg,[1,numel(F)]);

% iterative diagonalization
Inrg = NRG_IterDiagQS(H0,A0,Lambda,ff,F,gg,NF,Z,Nkeep);
%% 
% Plot the energy flow diagram.

% Energy flow diagram
Eshow = 3; % energy window to show (from 0 to Eshow)

% determine energy levels, corresponding quantum numbers and degeneracies

Es = cell(numel(Inrg.HK),1); % energies
Qs = cell(numel(Inrg.HK),1); % corresponding quantum numbers

for itN = 1:numel(Es)
    if itN < N
        HN = Inrg.HK{itN}; % kept Hamiltonian at itN
    else
        HN = Inrg.HD{itN}; % only discarded states at the last iteration
    end
    
    if itN == 1 % impurity is not diagonalized in NRG_IterDiagQS
        [~,I] = eigQS(HN);
        HN = QSpace(I.EK);
    end
    
    for itq = 1:size(HN.Q{1},1)
        Qs{itN}(itq,:) = HN.Q{1}(itq,:);
        Es{itN}{itq,1} = sort((HN.data{itq}(HN.data{itq} <= Eshow))','ascend');
    end
end

% get maximum number of energy levels per symmetry sector
nEmax = 0;
for iti = 1:numel(Es)
	nEmax = max([nEmax;cellfun('prodofsize',Es{iti}(:))]);
end

Quniq = unique(cell2mat(Qs),'rows'); % unique sets of quantum numbers
Es2 = cell(size(Quniq,1),1); % each Es2{n} is the collection of curves for each quantum number set Quniq(n,:)
Eeven = cell(size(Quniq,1),1); % even iterations
Eodd = cell(size(Quniq,1),1); % even iterations

for itq = (1:size(Quniq,1))
    Es2{itq} = nan(numel(Es)+1,nEmax);
    
    for iti = 1:numel(Es)
        okq = ismember(Qs{iti},Quniq(itq,:),'rows');
        
        if any(okq)
            Etmp = Es{iti}{okq};
            Es2{itq}(iti,(1:numel(Etmp))) = Etmp(:).';
        end
    end
    
    Es2{itq}(:,(find(~all(isnan(Es2{itq}),1),1,'last')+1:end)) = [];
    
    Eeven{itq} = Es2{itq}(1:2:end,:); % impurity is 0, i.e. even
    Eodd{itq} = Es2{itq}(2:2:end,:);
    
end

% shorten Eeven and Eodd such that only a single row of NaN's remains
Neven = 0; Nodd = 0;
for itq = 1:numel(Eeven)
    Neven = max([Neven,find(~all(isnan(Eeven{itq}),2),1,'last')+1]); 
    Nodd = max([Nodd,find(~all(isnan(Eodd{itq}),2),1,'last')+1]);
end
for itq = 1:numel(Eeven)
    
    if size(Eeven{itq},1) >= Neven
        Eeven{itq} = Eeven{itq}(1:Neven,:);
    else
        Eeven{itq}(end+1,:) = nan;
    end
    
    if size(Eodd{itq},1) >= Nodd
        Eodd{itq} = Eodd{itq}(1:Nodd,:);
    else
        Eodd{itq}(end+1,:) = nan;
    end
    
end

% quantum numbers for even and odd iterations
Qeven = Quniq;
Qodd = Quniq;

% delete all NaN sectors
iQeven = cellfun(@(x) all(isnan(x(:))),Eeven);
Eeven = Eeven(~iQeven);
Qeven = Qeven(~iQeven,:);

iQodd = cellfun(@(x) all(isnan(x(:))),Eodd);
Eodd = Eodd(~iQodd);
Qodd = Qodd(~iQodd,:);

% sort by lowest energy at last iteration
[~,sid] = sort(cellfun(@(x) min(x(end-1,:)), Eeven),'ascend'); % for 'min' and 'sort', NaN is considered to be the largest
Qeven = Qeven(sid,:);
Eeven = Eeven(sid);

[~,sid] = sort(cellfun(@(x) min(x(end-1,:)), Eodd),'ascend'); % for 'min' and 'sort', NaN is considered to be the largest
Qodd = Qodd(sid,:);
Eodd = Eodd(sid);

% colors:
clrs = [228,26,28;...
    55,126,184;...
    77,175,74;...
    152,78,163;...
    255,127,0]/256; % colors for lowest few energy levels
                    % size(clrs,1) determines number of colored levels
% default line color: gray
clrs0 = [1 1 1]*0.7;

% plot flow diagram:
figure;
% upper panel
subplot(2,1,1);
hold on;
for itq = (size(clrs,1)+1):size(Qeven,1) % first gray lines
    plot(2*((1:size(Eeven{itq},1))-1),Eeven{itq},'LineWidth',1,'Color',clrs0);
end
hs = cell(1,size(clrs,1)); % lines
leg = cell(1,size(clrs,1)); % legend
for itq = 1:size(clrs,1) % first gray lines
    hs{itq} = plot(2*((1:size(Eeven{itq},1))-1),Eeven{itq},'LineWidth',1,'Color',clrs(itq,:));
    leg{itq} = ['(',...
        sprintf([repmat(['%i '],[1,size(Qeven(itq,:),2)-1]),'%i'],Qeven(itq,:))...
        ,')'];
end
hs2 = [];
for it1 = (1:numel(hs)) % for legend
    hs2(it1) = hs{it1}(1);
end
legend(hs2,leg);
xlabel('Even iterations');
xlim([0 numel(Inrg.HK)-1]);
ylim([0, Eshow]);
set(gca,'LineWidth',1,'FontSize',13);
hold off;

% lower panel
subplot(2,1,2);
hold on;
for itq = (size(clrs,1)+1):size(Qodd,1) % first gray lines
    plot(2*(1:size(Eodd{itq},1))-1,Eodd{itq},'LineWidth',1,'Color',clrs0);
end
hs = cell(1,size(clrs,1)); % lines
leg = cell(1,size(clrs,1)); % legend
for itq = 1:size(clrs,1) % first gray lines
    hs{itq} = plot(2*(1:size(Eodd{itq},1))-1,Eodd{itq},'LineWidth',1,'Color',clrs(itq,:));
    leg{itq} = ['(',...
        sprintf([repmat(['%i '],[1,size(Qodd(itq,:),2)-1]),'%i'],Qodd(itq,:))...
        ,')'];
end
hs2 = [];
for it1 = (1:numel(hs)) % for legend
    hs2(it1) = hs{it1}(1);
end
legend(hs2,leg);
xlabel('Odd iterations');
xlim([0 numel(Inrg.HK)-1]);
ylim([0, Eshow]);
set(gca,'LineWidth',1,'FontSize',13);
hold off;
%% Exercise 1: Complete |getRhoFDMQS_Ex.m|
% There is a function |getRhoFDMQS_Ex.m| which is zipped together with this 
% document. *Complete the parts which are enclosed by the comments |TODO - Exercise 
% 1*|.
% 
% Below is the demonstration in which the completed function is used. In the 
% next section ("Double occupation vs. temperature"), the reduced density matrix 
% at the impurity, which is the part of the output (|Inrg.RK{1}|) of the function 
% |getRhoFDMQS_Ex.m|, is computed and displayed for different temperatures. Compare 
% the result from your version of function with the values shown below.
%% Double occupation vs. temperature
% Let's compute the double occupation $\langle \hat{n}_{d \uparrow} \hat{n}_{d\downarrow} 
% \rangle$ at the impurity, where $\hat{n}_{d\sigma}$ is the number operator for 
% spin-$\sigma$ electron at the impurity site. We obtain the reduced density matrix 
% within the Hilbert space of the impurity (to be obtained as |Inrg.RK{1}|), where 
% the bath sites are all traced out.

Dop = getsub(EF,[1 0],1); % operator measuring for double occupation
% represent in kept basis at the impurity
Dop = contract(Inrg.AK{1},'!2*',{Dop,Inrg.AK{1}});

T = 10.^(1:-1:-7); % temperature values
Dval = zeros(size(T)); % expectation value of Dop

for itT = (1:numel(T))
    Inrg = getRhoFDMQS_Ex(Inrg,T(itT)); % construct FDM
    % display the reduced density matrix at the impurity
    Inrg.RK{1}
    % contract the reduced density matrix at the impurity 
    % with the operator
    Dval(itT) = trace(contract(Inrg.RK{1},Dop));
end
%% 
% Plot the double occupation vs. temperature.

figure;
semilogx(T/U,Dval,'Marker','x','LineWidth',1);
set(gca,'FontSize',13,'LineWidth',1);
xlabel('$T/U$','Interpreter','latex');
ylabel('$\langle \hat{n}_{d \uparrow} \hat{n}_{d \downarrow} \rangle$', ...
    'Interpreter','latex');
grid on;
%% 
% We see that the double occupation $\langle \hat{n}_{d \uparrow} \hat{n}_{d 
% \downarrow} \rangle$ decreases with decreasing $T$ with a crossover around $T 
% \sim U$. On the other hand, the double occupation saturates in the low-temperature 
% regime $T \ll U$: it means that the charge fluctuation is "frozen" by low temperature.
%% Exercise 2: Complete |getAdisc_Ex.m|
% There is a function |getAdisc_Ex.m| which is zipped together with this document. 
% *Complete the parts which are enclosed by the comments |TODO - Exercise 2*|.
% 
% (_Hint_: The function can be completed by adding only three lines!)
%% Impurity spectral function: discrete data
% We compute the spectral function $A_{d s} (\omega) \equiv (-1/\pi) \mathrm{Im} 
% \, G_{d s} (\omega)$, where $G_{d s} (\omega)$ is the Green's function of spin-$s$ 
% electron at the impurity. That is, $A_{d s} (\omega)$ is obtained by substituting 
% $\hat{O}_d = \hat{d}_s$ to the above Eq. (1).
% 
% Choose the system temperature, and construct the density matrices in the kept 
% and discarded spaces for every iteration.

T = 1e-8;
Inrg = getRhoFDMQS_Ex(Inrg,T); % construct full density matrix
%% 
% Then compute the discrete spectal data for the impurity spectral function. 

% discrete spectral function of spin-up electron at the impurity
[odisc,Adisc] = getAdiscQS_Ex(Inrg,F(1),Z);
%% 
% Note that |sum(Adisc)| means the sum of the discrete spectral weights. The 
% sum of the spectral weights is given by
% 
% $$\sum_\ell \sum_{X, Y \neq K K} \sum_{\alpha \beta} A_{\alpha \beta}^{\ell 
% X Y}$$
% 
% $$= \sum_{\ell} \sum_{X,Y \neq KK} \sum_{\alpha \beta} \sum_e {}_\ell^X \langle 
% \alpha, e | \hat{O}_d | \beta, e \rangle_\ell^Y\left( {}_\ell^Y \langle \beta, 
% e | \hat{O}_d^\dagger \, \hat{\rho} | \alpha, e \rangle_\ell^X\pm{}_\ell^Y \langle 
% \beta, e | \hat{\rho} \, \hat{O}_d^\dagger | \alpha, e \rangle_\ell^X \right)$$
% 
% $$= \mathrm{Tr} \left( \hat{O}_d \hat{O}_d^\dagger \hat{\rho} \pm \hat{O}_d 
% \hat{\rho} \hat{O}_d^\dagger \right) = \mathrm{Tr} \left( [\hat{O}_d, \hat{O}_d^\dagger 
% ]_\pm \hat{\rho} \right) .$$
% 
% Here we have used the completeness of the basis $\{ | \alpha, e \rangle_{\ell_0}^D 
% \} \cup \{ | \alpha, e \rangle_{\ell_0 + 1}^D \} \cup \cdots \{ | \alpha, e 
% \rangle_{\ell}^D \} \cup \{ | \alpha, e \rangle_{\ell}^K \}$, where $\ell_0$ 
% is the first iteration at which the truncation of the Hilbert space happened. 
% Since the standard anti-commutation relation gives $[\hat{d}_s, \hat{d}_s^\dagger]_+ 
% = 1$, |*sum(Adisc)| should be 1. The |getAdiscQS_Ex| result satisfies this "spectral 
% sum rule" up to double precision.* Indeed, this sum-preserving property is one 
% of the key merits of fdmNRG; see the title of [<https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.99.076402 
% A. Weichselbaum and J. von Delft, Phys. Rev. Lett. *99*, 076402 (2007)>]. Note 
% that in case of SU(2) spin rotation symmetry, we compute $A_{\uparrow}(\omega) 
% + A_{\downarrow}(\omega)$. Therefore, sum(Adisc) is 2 in this case. 
% 
% We use a technique called "rerouting Z-string" in |getAdiscQS_Ex.m|. With 
% this, we don't need to contract Z operator, i.e., fermionic sign operator, at 
% every iteration. It is enough to contract the Z operator only at the first iteration! 
% For the detail of the rerouting technique, see the Appendix of [<https://journals.aps.org/prb/abstract/10.1103/PhysRevB.86.245124 
% A. Weichselbaum, Phys. Rev. B *86*, 245124 (2012)>] or <https://arxiv.org/abs/1209.2062 
% its arXiv version>.
%% Impurity spectral function: continuous curve
% The results |odisc| and |Adisc| represent the collection of discrete spectral 
% weights given by the Lehmann representation. To obtain a smooth curve, we need 
% to broaden them.

% broaden the discrete data to have a continuous curve
[ocont,Acont] = getAcont(odisc,Adisc,log(Lambda),T/5);
if strcmp(symstr,'Acharge,SU2spin')
    Acont = Acont/2; % sum of spectral function for both spins is calculated
end
%% 
% Compute the Kondo temperature $T_\mathrm{K}$ to rescale frequencies.

% Kondo temperature from the Bethe ansatz
TK = sqrt(U*Gamma/2)*exp(-pi*U/8/Gamma + pi*Gamma/2/U);
disp(TK); % Kondo temperature
figure;
hold on;
% positive frequency
plot(ocont(ocont>0)/TK,Acont(ocont>0)*(pi*Gamma), ...
    'LineWidth',1,'LineStyle','-');
% negative frequency
plot(-ocont(ocont<0)/TK,Acont(ocont<0)*(pi*Gamma), ...
    'LineWidth',1,'LineStyle','--');
plot(U/2/TK*[1 1],[0 1],'LineWidth',1,'LineStyle','-.', ...
    'Color',[1 1 1]*0.5);
text(U/TK,0.3,'$U/2/T_\mathrm{K}$','FontSize',13, ...
    'Interpreter','latex','HorizontalAlignment','left');
hold off;
set(gca,'FontSize',13,'LineWidth',1,'XScale','log');
legend({'$A_{d\uparrow} (\omega > 0)$', ...
    '$A_{d\uparrow} (\omega < 0)$'},'Interpreter','latex');
xlabel('$| \omega | / T_\mathrm{K}$','Interpreter','latex');
ylabel('$\pi \Gamma A(\omega)$','Interpreter','latex');
ylim([0 1.01]);
xlim([1e-6 1e6]);
grid on;
%% 
% We observe several important features.
%% 
% # As the spin symmetry is not broken (due to the absence of magnetic field) 
% and the particle-hole symmetry is preserved, blue and red curves for positive 
% and negative frequencies, respectively, lie top of each other.
% # There are side peaks at $\omega = \pm U/2$, as marked by vertical dash-dotted 
% line. They come from local excitations from singly occupied states to the double 
% occupied state for the peak at $\omega = U/2$ (to the empty state for the peak 
% at $\omega = -U/2$).
% # The full-width-half-maximum of the "Kondo peak" (peak centered at $\omega 
% = 0$) is roughly $T_\mathrm{K}$.
% # There is a plateau for $| \omega | \ll T_\mathrm{K}$ which corresponds to 
% the top of the Kondo peak. The height of the Kondo peak satisfies the Friedel 
% sum rule, $\pi \Gamma A_{d\uparrow} (\omega = 0) = 1$, up to a few percent error. 
% Note that the Friedel sum rule is the exact relation holds at $T = 0$.
%% 
% The quality of the curve (e.g., frequency resolution) can be improved by using 
% so-called $z$-averaging, that is, by mixing the NRG results with the logarithmic 
% discretization grid relatively shifted with respect to each other.
% 
% Also, the accuracy with respect to the Friedel sum rule can be further improved 
% by using more advanced logarithmic discretization schemes (see e.g., [<https://journals.aps.org/prb/abstract/10.1103/PhysRevB.79.085106 
% R. Zitko and Th. Pruschke, Phys. Rev. B *79*, 085106 (2009)>]) and by using 
% the "self-energy trick" (see [<https://iopscience.iop.org/article/10.1088/0953-8984/10/37/021/meta 
% R. Bulla, A. C. Hewson, and Th. Pruschke, J. Phys.: Condens. Matter *10*, 8365 
% (1998)>]).
%% Exercise 3: Dynamical impurity spin susceptibility
% In the accompanying tutorial T11a, we have computed the impurity contribution 
% to the spin susceptibility. This property is static, i.e., does not depend on 
% frequency $\omega$. Some experimental probes such as neutron scattering can 
% resolve the frequency dependence of the spin susceptibility. The frequency dependence 
% encodes interesting information of the system.
% 
% In this exercise, we will consider the dynamical impurity spin susceptibility, 
% which depends on frequency $\omega$. It is the retarded-time correlation function 
% of spin operators $\hat{S}_{d,z}$. That is, the imaginary part of the dynamical 
% impurity spin susceptibility, say $\chi''_s (\omega)$, is obtained by substituting 
% $\hat{O}_d = \hat{S}_{d,z}$ to the above Eq. (1).
% 
% Compute $\chi''_s (\omega)$ for the SIAM with the same choice of parameters 
% as in the above demonstration of the impurity spectral function. How does the 
% curve $\chi''_s (\omega)$ look like? Can you interpret the location of the peak 
% of the curve $\chi''_s (\omega)$?
% 
% (_Hint 1_: Use |getAdiscQS_Ex| by setting the third input |Z| as |[]| (empty).)
% 
% (_Hint 2_: $\chi''_s(\omega)$ is an odd function, having peaks as $\omega 
% \simeq \pm T_\mathrm{K}$. Refer to Fig. 1 of [<https://journals.aps.org/prb/abstract/10.1103/PhysRevB.89.075130 
% M. Hanl and A. Weichselbaum, Phys. Rev. B *89*, 075130 (2014)>] for the shape 
% of $\chi''_s(\omega)$ for $\omega > 0$.)